@@ -5,48 +5,45 @@ use Carp qw(carp croak);
 use File::Spec;
 use FileHandle;
 
-use vars qw($CLASS $VERSION);
-$CLASS   = 'Config::Natural';
-$VERSION = '0.98';
-
-my @base = (
-    options => {
-        'comment_line_symbol'     => '#', 
-        'affectation_symbol'      => '=', 
-        'multiline_begin_symbol'  => '-', 
-        'multiline_end_symbol'    => '.', 
-        'list_begin_symbol'       => '{', 
-        'list_end_symbol'         => '}', 
-        'include_symbol'          => 'include', 
-        'case_sensitive'          => 1, 
-        'auto_create_surrounding_list' => 1, 
-        'read_hidden_files'       => 0, 
-    }, 
-    state => {  }, 
-    param => {  }, 
-    handlers => {  }, 
-    prefilter => 0, 
-    filter => 0, 
-);
+{ no strict; 
+  $CLASS   = 'Config::Natural';
+  $VERSION = '0.99';
+}
 
 # class option
 my %options = (
     quiet => 0, 
 );
 
-## set the accessors for the object options
-for my $option (keys %{$base[1]}) {
-    eval qq| sub $option { _get_set_option(shift, '$option', shift) } |;
-    croak "Initialisation error: $@ " if $@;
-}
-
 
 # 
 # new()
 # ---
 sub new {
     my $class = shift;
-    my $self = bless { @base }, $class;
+    my $self = bless {
+        options => {
+            'comment_line_symbol'     => '#', 
+            'affectation_symbol'      => '=', 
+            'multiline_begin_symbol'  => '-', 
+            'multiline_end_symbol'    => '.', 
+            'array_begin_symbol'      => '(', 
+            'array_end_symbol'        => ')', 
+            'list_begin_symbol'       => '{', 
+            'list_end_symbol'         => '}', 
+            'include_symbol'          => 'include', 
+            'case_sensitive'          => 1, 
+            'auto_create_surrounding_list' => 0, 
+            'read_hidden_files'       => 0, 
+            'strip_indentation'       => 0, 
+        }, 
+        state => {  }, 
+        param => {  }, 
+        handlers => {  }, 
+        prefilter => 0, 
+        filter => 0, 
+    }, $class;
+    
     if(ref $_[0] eq 'HASH') {
         my $opts = shift;
         for my $option (keys %$opts) {
@@ -61,6 +58,34 @@ sub new {
 }
 
 
+# 
+# AUTOLOAD()
+# --------
+sub AUTOLOAD {
+    no strict;
+    my $self = $_[0];
+    my $type = ref $self || croak "I am not an object, so don't call me that way.";
+    my $name = $AUTOLOAD;
+    $name =~ s/.*:://;
+    
+    carp "Unknown option '$name'" 
+      unless exists $options{$name} or defined $self->{options}{$name};
+
+    my $code = q{
+        sub {
+            my $self = shift;
+            my $value = $self->{options}{METHOD};
+            $self->{options}{METHOD} = shift if @_;
+            return $value
+        }
+    };
+    $code =~ s/METHOD/$name/g;
+
+    *$AUTOLOAD = eval $code;
+    goto &$AUTOLOAD;
+}
+
+
 # 
 # DESTROY()
 # -------
@@ -71,25 +96,6 @@ sub DESTROY {
 }
 
 
-# 
-# _get_set_option()
-# ---------------
-sub _get_set_option {
-    my $self   = shift;
-    my $option = shift;
-    my $value  = shift;
-    
-    carp "Unknown option '$option' " unless exists $self->{options}{$option} or $options{'quiet'};
-    
-    if(defined $value) {
-        ($value, $self->{options}{$option}) = ($self->{options}{$option}, $value);
-        return $value
-    } else {
-        return $self->{options}{$option}
-    }
-}
-
-
 # 
 # options()
 # -------
@@ -99,13 +105,13 @@ sub options {
     my @ret_list = ();
     
     for my $arg (@{$args->{'get'}}) {
-        carp "Class option '$arg' does not exist " and next 
+        carp "Class option '$arg' does not exist" and next 
             unless exists $options{$arg} or $options{'quiet'};
         push @ret_list, $options{$arg};
     }
     
     for my $arg (keys %{$args->{'set'}}) {
-        carp "Class option '$arg' does not exist " and next 
+        carp "Class option '$arg' does not exist" and next 
             unless exists $options{$arg} or $options{'quiet'};
         $options{$arg} = $args->{'set'}{$arg};
     }
@@ -167,13 +173,15 @@ sub read_source {
     my $aff_sym   = $self->affectation_symbol;
     my $multiline = $self->multiline_begin_symbol;
     my $multi_end = $self->multiline_end_symbol;
+    my $array     = $self->array_begin_symbol;
+    my $array_end = $self->array_end_symbol;
     my $list      = $self->list_begin_symbol;
     my $list_end  = $self->list_end_symbol;
     my $include   = $self->include_symbol;
     my $state     = $self->{'state'};
     
     # store the name of the last opened file
-    $state->{filename} = $file;
+    $state->{'filename'} = $file;
     
     while(defined($_ = <$fh>)) {
         ## execute the prefilter if present
@@ -186,7 +194,7 @@ sub read_source {
         ## include statement
         if(/^\s*\Q${include}\E\s+(\S+)\s*$/) {
             my $included = $1;
-            my @path = File::Spec->splitdir($state->{filename});
+            my @path = File::Spec->splitdir($state->{'filename'});
             pop @path;  # remove the current file name from the path
             $included = File::Spec->catdir(@path, $included);
             $self->read_source($included);
@@ -223,12 +231,30 @@ sub read_source {
             $value = '';
             $_ = <$fh>;
             
+            $self->strip_indentation and my($indent) = (/^(\s*)/);
+            
             while(not /^\s*\Q${multi_end}\E\s*$/) {
+                $indent and s/^$indent//;
                 $value .= $_;
                 $_ = <$fh>;
             }
         }
         
+        ## array case
+        if($value =~ /^\s*\Q${array}\E\s*$/) {
+            $value = '';
+            $_ = <$fh>;
+            
+            while(not /^\s*\Q${array_end}\E\s*$/) {
+                s/^\s*//;
+                $value .= $_;
+                $_ = <$fh>;
+            }
+            
+            $self->param({ $field => [ split $/, $value ] });
+            next
+        }
+        
         ## create a surrounding list if the parameter already exists
         if($self->auto_create_surrounding_list) {
             my $surrlist   = "${field}s";
@@ -287,7 +313,7 @@ sub param {
     
     ## get the value of the desired parameters
     for my $arg (@{$args->{'get'}}) {
-        carp "Parameter '$arg' does not exist " and next
+        carp "Parameter '$arg' does not exist" and next
             if not exists $self->{'param'}{_case_($self, $arg)} and not $options{'quiet'};
         
         push @retlist, $self->{'param'}{_case_($self, $arg)}
@@ -352,7 +378,7 @@ sub _parse_args {
                 }
                 
             } else {
-                carp "Bad ref $ref_type; ignoring it " unless $options{'quiet'};
+                carp "Bad ref $ref_type; ignoring it" unless $options{'quiet'};
                 next
             }
         
@@ -361,7 +387,7 @@ sub _parse_args {
            if(substr($arg, 0, 1) eq '-') {
                $arg = substr($arg, 1);
                my $val = shift;
-               carp "Undefined value for parameter '$arg' " and next 
+               carp "Undefined value for parameter '$arg'" and next 
                    if not defined $val and not $options{'quiet'};
                $args{'set'}{$arg} = $val if $arg
                
@@ -463,6 +489,52 @@ sub all_parameters {
 }
 
 
+# 
+# param_tree()
+# --------------
+# Return the hash tree of all parameters
+# 
+sub param_tree {
+    my $self = shift;
+    my $tree = {};
+    $tree->{$_} = $self->param($_) for($self->param());
+    return $tree;
+}
+
+
+# 
+# value_of()
+# --------
+# Return the value of the specified parameter
+# 
+sub value_of {
+    my $self = shift;
+    my $param_path = shift;
+    
+    # handle simple cases simply...
+    return $self->{'param'}{$param_path} if $self->{'param'}{$param_path};
+    
+    # handle more complex cases nicely.
+    my @path = split '/', $param_path;
+    not $path[0] and shift @path;
+    
+    my($name,$index) = ( (shift @path) =~ /^([^[]+)(?:\[([+-]?\d+|\*)\])?$/ );
+    my $node = $self->param($name);  $index ||= 0;
+    return $node if $index eq '*';
+    
+    if(ref $node) {
+        $node = $node->[int($index)];
+        for my $p (@path) {
+            ($name,$index) = ( ($p) =~ /^([^[]+)(?:\[([+-]?\d+|\*)\])?$/ );
+            $node = $node->{$name};  $index ||= 0;
+            ref $node and $index ne '*' and $node = $node->[int($index)];
+        }
+    }
+    
+    return $node
+}
+
+
 # 
 # delete()
 # ------
@@ -472,7 +544,7 @@ sub delete {
     my $self = shift;
     
     for my $param (@_) {
-        carp "Parameter '$param' does not exist " and next 
+        carp "Parameter '$param' does not exist" and next 
             if not exists $self->{'param'}{_case_($self, $param)} and not $options{'quiet'};
         delete $self->{'param'}{_case_($self, $param)}
     }
@@ -534,9 +606,9 @@ sub _dump_tree {
     my $level = shift;
     my %state = @_;
     my $str = '';
+    my $sp = $state{'nospace'} ? '' : ' ';
     
     if(ref $tree eq 'HASH') {
-        my $sp = $state{'nospace'} ? '' : ' ';
         
         # add the list name and symbol
         $state{'list_name'} and 
@@ -568,7 +640,16 @@ sub _dump_tree {
                 $sp x (($level-1)*2), $self->list_end_symbol, $/;
     
     } elsif(ref $tree eq 'ARRAY') {
-        for my $list (@$tree) { $str .= _dump_tree($self, $list, $level, %state) }
+        if(ref $tree->[0]) {
+            for my $list (@$tree) { $str .= _dump_tree($self, $list, $level, %state) }
+        } else {
+            $str .= join '', 
+                    $sp x ($level*2), 
+                    $state{'prefix'}, $state{'list_name'}, $sp,
+                    $self-> affectation_symbol, $sp, $self->array_begin_symbol, $/, 
+                    (map {($sp x (($level+1)*2)) . $_ . $/} @$tree), 
+                    $sp x ($level*2), $self->array_end_symbol, $/
+        }
     
     } else {
         warn "unexpected reference type ", ref($tree)
@@ -641,18 +722,44 @@ programs as simple as:
     use strict;
     use Config::Natural;
     use HTML::Template;
-    
+
     my $source = new Config::Natural 'file.src';
     my $tmpl = new HTML::Template type => 'filename', 
             source => 'file.tmpl', associate => $source;
     print $tmpl->output;
 
 And this is not just another trivial example: I use scripts nearly 
-as simple as this one to create most of my pages. 
+as simple as this one to create most of my web pages. 
 
 
 =head1 SYNTAX
 
+In brief, Config::Natural supports the following features in 
+configuration files: 
+
+=over 4
+
+=item *
+
+multiline values, 
+
+=item *
+
+arrays, 
+
+=item *
+
+nested lists, 
+
+=item *
+
+inclusion. 
+
+=back
+
+Any element of the syntax can be changed via the corresponding accessor. 
+
+
 =head2 Affectations
 
 To affect a value to a parameter, simply write:
@@ -675,9 +782,22 @@ You can also give multi-lines values this way:
 
 Think of this as a "Unix inspired" syntax. Instead of giving the value, 
 you write C<"-"> to mean "the value will follow" (in Unix, this means the 
-data will come from standard input). To end the multi-lines value, you 
-simply put a single dot C<"."> on a line (as in Unix mail, but it needn't 
-be on the first column). 
+data will come from standard input). To end the multi-lines value, put 
+a single dot C<"."> on a line (as in Unix mail, but it needn't be on 
+the first column). 
+
+
+=head2 Arrays
+
+Since version 0.99, Config::Natural also supports arrays, using the 
+following syntax: 
+
+    fruits = (
+        apple
+        banana
+        kiwi
+        orange
+    )
 
 
 =head2 Lists
@@ -701,13 +821,13 @@ Example: a version history
         version = 0.7.0
         comment = First fully functional release as an independent module.
     }
-    
+
     history {
         date = 2000.11.04
         version = 0.7.1
         comment = Minor change in the internal structure: options are now grouped.
     }
-    
+
     history {
         date = 2000.11.05
         version = 0.8.0
@@ -720,7 +840,7 @@ Lists can be nested. Example:
     machine {
         name = neutron
         sys = linux
-        
+
         service {
             type = firewall
             importance = critical
@@ -730,26 +850,28 @@ Lists can be nested. Example:
     machine {
         name = proton
         sys = linux
-        
+
         service {
             type = proxy
             importance = low
-       }
-        
+        }
+
         service {
             type = router
             importance = medium
        }
     }
 
+Note that list blocks need not to contain exactly the same set of parameters
 
-As a shorthand, you can write something like
+If you enable the C<auto_create_surrounding_list> option, you can then write 
+something like
 
     flavour = lemon
     flavour = strawberry
     flavour = vanilla
 
-instead of
+as a shorthand for 
 
     flavours {
         flavour = lemon
@@ -840,7 +962,7 @@ used in order to correct some names which otherwise couldn't be parsed
 by Config::Natural, for example names with spaces. Check in the 
 F<examples/> directory for sample programs that implements such functions. 
 
-You can set up a prefilter using the C<-E<gt>prefilter()> method, or at 
+You can set up a prefilter using the C<prefilter()> method, or at 
 creation time with the C<prefilter> option. 
 
 =head2 Data filter
@@ -850,7 +972,7 @@ This can be used to implement additional features, like a syntax
 for interpolating values. Check in the F<examples/> directory for 
 sample programs that implements such functions. 
 
-You can set up a data filter using the C<-E<gt>filter()> method, or at 
+You can set up a data filter using the C<filter()> method, or at 
 creation time with the C<filter> option. 
 
 
@@ -861,7 +983,57 @@ but instead of being object methods, handlers can be seen as
 "parameters" methods, in that they are bound to a name, and are 
 only called when a parameter with that name is affected. 
 
-Handlers are defined with the C<-E<gt>handler()> method. 
+Handlers are defined with the C<set_handler()> method. 
+
+
+=head1 PARAMETER PATH
+
+This is a new functionality, introduced in version 0.99. 
+
+A parameter path is a way of referring any parameter, even if 
+it's deeply buried inside several layers of nested lists. 
+It is used by the method C<value_of()> to provide a much 
+easier way to read data hidden in nested lists. 
+
+The parameter path syntax is loosely inspired by XPath:
+
+    path = /level0[index0]/level1[index1]/.../param
+
+Indexes start at zero, like in Perl (and unlike XPath). 
+When an index is omitted, C<[0]> is assumed. 
+
+Examples: 
+
+    # same as $config->param('myparam')
+    $value = $config->value_of('/myparam');
+
+    # same as $config->param('list')->[0]{myparam}
+    $value = $config->value_of(/list[0]/myparam);
+    $value = $config->value_of(/list/myparam);
+
+If you want to get back a whole list, instead of a single value, 
+use C<[*]> as the last index, and it will return the reference to 
+that list. 
+
+    # same as $config->param('list')
+    $value = $config->value_of('/list[*]');
+
+    # same as $config->param('list')->[0]{inner_list}
+    $value = $config->value_of('/list/inner_list[*]');
+
+This syntax also applies to arrays, so when you read
+
+    operators = (
+        Ibuki Maya
+        Hyuga Makoto
+        Aoba Shigeru
+    )
+
+then the following code behaves as expected: 
+
+    print $nerv->value_of("/operators");     # prints "Ibuki Maya"
+    print $nerv->value_of("/operators[1]");  # prints "Hyuga Makoto"
+    $ref = $nerv->value_of("/operators[*]"); # return the corresponding arrayref. 
 
 
 =head1 OBJECTS OPTIONS
@@ -870,7 +1042,8 @@ Handlers are defined with the C<-E<gt>handler()> method.
 
 If the default symbols used in the configuration file syntax doesn't 
 fit your needs, you can change them using the following methods. 
-Of course you must call these I<before> reading the data file(s). 
+Of course you must call these I<before> reading the configuration 
+files. 
 
 =over 4
 
@@ -889,6 +1062,16 @@ Use this accessor to change the multiline begin symbol. Default is C<"-">.
 Use this accessor to change the multiline end symbol. Default is C<".">.
 
 
+=item array_begin_symbol
+
+Use this accessor to change the array begin symbol. Default is C<"(">.
+
+
+=item array_end_symbol
+
+Use this accessor to change the array end symbol. Default is C<")">.
+
+
 =item comment_line_symbol
 
 Use this accessor to change the comment symbol. Default is C<"#">.
@@ -915,15 +1098,15 @@ Use this accessor to change the include symbol. Default is C<"include">.
 
 =over 4
 
-=item case_sensitive
-
-Use this accessor to change the case behaviour. Default is 1 (case sensitive). 
-
-
 =item auto_create_surrounding_list
 
 Use this accessor to enable or disable the auto creation of surrounding 
-lists. Default is 1 (enabled). 
+lists. Default is 0 (disabled). 
+
+
+=item case_sensitive
+
+Use this accessor to change the case behaviour. Default is 1 (case sensitive). 
 
 
 =item read_hidden_files
@@ -931,6 +1114,29 @@ lists. Default is 1 (enabled).
 Use this accessor to allow of forbid Config::Natural to read hidden files 
 when reading a directory. Default is 0 (don't read hidden files). 
 
+
+=item strip_indentation
+
+Use this accessor to enable Config::Natural to automatically strip the 
+indentation from multilines values. For example, when this option is 
+enabled, the following: 
+
+    text = -
+        I thought what I'd do was. 
+        I'd pretend I was one of those deaf-mutes or should I?
+            --Laughing Man
+    .
+
+will be converted in
+    
+    text = -
+I thought what I'd do was. 
+I'd pretend I was one of those deaf-mutes or should I?
+    --Laughing Man
+    .
+
+This option is disabled by default. 
+
 =back
 
 
@@ -1040,6 +1246,19 @@ the value of the parameters of an object.
     );
 
 
+=item param_tree ( )
+
+This method returns a hashref that allows direct access to the tree 
+of parameters. 
+
+
+=item value_of ( I<PARAMETER_PATH> )
+
+This method is an easier way to access the values of the parameters. 
+It returns the value of the parameter path given in argument. 
+Check L<"PARAMETER PATH"> for more information and some examples.
+
+
 =item all_parameters ( )
 
 This method returns the list of the parameters of an object.
@@ -1070,7 +1289,9 @@ This method sets all the parameters to undef.
 This method returns a dump of the parameters as a string using the 
 current format of the Config::Natural object. It can be used to simply 
 print them out, or to save them to a configuration file which can be 
-re-read by another Config::Natural object.
+re-read by another Config::Natural object. 
+
+Options are passed as a hashref. 
 
 B<Options>
 
@@ -1101,7 +1322,7 @@ after printing each parameter.
 =item write_source ( I<FILEHANDLE> [, I<OPTIONS>] )
 
 This method writes the current object to the given file name or file 
-handle. Remaining parameters, if any, will be passed unmodified to 
+handle. Remaining options, if any, will be passed unmodified to 
 dump_param(). If no argument is given, the file or handle used by 
 the last call of read_source() will be used. 
 
@@ -1120,7 +1341,7 @@ For example:
         $data =~ s/\s*#.*$//go;  # remove comments appearing on 
         return $data          # an affectation line
     }
-    
+
     my $conf = new Config::Natural { filter => \&myfilter };
 
 
@@ -1144,7 +1365,7 @@ be used as the actual value. An example function could look like this:
         my $value = shift;
         return crypt($value, $param);
     }
-    
+
     $conf->set_handler('passwd', \&crypt_handler);
 
 
@@ -1167,7 +1388,7 @@ SE<eacute>bastien Aperghis-Tramoni <sebastien@aperghis.net>
 
 =head1 COPYRIGHT
 
-Config::Natural is Copyright (C)2000-2003 SE<eacute>bastien Aperghis-Tramoni.
+Config::Natural is Copyright (C)2000-2004 SE<eacute>bastien Aperghis-Tramoni.
 
 This program is free software. You can redistribute it and/or modify it 
 under the same terms as Perl itself. 
